""" Tests for gui model.
"""
import os
import tempfile
import shutil
import unittest
from pubsub import pub
from unittest import mock
from kvalikirstu2 import argument_parser
from kvalikirstu2 import exceptions
from kvalikirstu2 import gui_model
from kvalikirstu2 import kvali_interface
from tests import test_writer

test_files_path = os.path.join(os.path.dirname(__file__), "test files", "gui_model_test")


class TestGUIModel(unittest.TestCase):
    def setUp(self):
        pub.unsubAll()
        argument_parser.Settings().set_up_cli_settings(['--max-backup-folder-size', '0'])
        self.temp_dir = tempfile.TemporaryDirectory()
        self.temp_path = os.path.join(self.temp_dir.name, "test")
        shutil.copytree(test_files_path, self.temp_path)

        self.model = gui_model.GUIModel()
        self.model.set_study_path(self.temp_path)
        self.model.set_type_shown(gui_model.HeaderType.NOT_HEADER, True)
        self.model.set_type_shown(gui_model.HeaderType.BUILTIN, True)
        self.model._set_unsaved_changes(False)

    def tearDown(self):
        self.temp_dir.cleanup()

    def test_save_changes(self):
        self.model.set_row_name(1, "test")
        self.model.save_changes()
        self.model.set_study_path(self.temp_path)
        self.assertEqual(self.model.visible_headers[1].edited_header, "test")

    def test_set_row(self):
        self.model.set_row_name(1, "test")
        self.assertEqual(self.model.visible_headers[1].edited_header, "test")

    def test_change_header(self):
        self.model.change_header_property([1], lambda header: header.set_status(gui_model.HeaderStatus.HIDE))
        self.assertEqual(self.model.visible_headers[1].status, gui_model.HeaderStatus.HIDE)

    def test_header_info_is_deep_copied(self):
        self.assertTrue(self.model.header_info is not kvali_interface._DATA['header_info'])

    def test_add_text_to_dafs(self):
        self.model.add_text_to_dafs('1\n2')
        test_data_path = os.path.join(self.temp_path, "data", "test_data.txt")
        lines = test_writer.read_lines_from_file(test_data_path)
        self.assertEqual("1", lines[0])
        self.assertEqual("2", lines[1])

    def test_check_unsaved_changes(self):
        self.model.set_row_name(0, "test")
        self.assertTrue(self.model.check_unsaved_changes())
        self.model.save_changes()
        self.assertFalse(self.model.check_unsaved_changes())

    def test_check_unsaved_changes_daf(self):
        self.model.set_row_name(0, "test")
        self.assertTrue(self.model.check_unsaved_changes())
        self.model.save_changes_to_daf()
        self.assertFalse(self.model.check_unsaved_changes())

    def test_conversion(self):
        self.model.convert_to_odt()
        test_data_path = os.path.join(self.temp_path, "data", "to_convert.odt")
        self.assertTrue(os.path.exists(test_data_path))

    def test_delete_tempfiles(self):
        self.model.generate_tempfiles()
        self.assertTrue(kvali_interface.temp_files_exist())
        self.model.delete_tempfiles()
        self.assertFalse(kvali_interface.temp_files_exist())

    def test_generate_citation_requirement(self):
        self.model.generate_citation_requirement()
        test_data_path = os.path.join(self.temp_path, "data", "test_data.txt")
        lines = test_writer.read_lines_from_file(test_data_path)
        self.assertEqual("citreq", lines[0])

    def test_generate_output(self):
        self.model.generate_tempfiles()
        self.model.generate_output()
        index_path = os.path.join(self.temp_path, "index.html")
        self.assertTrue(os.path.exists(index_path))

    def test_get_headerline(self):
        self.assertEqual("NUMERO", self.model.get_headerline(1).original_header)

    def test_move_down(self):
        self.model.visible_headers[1].index = 1000
        self.model.visible_headers[2].index = 2000
        self.assertTrue("SUKUPUOLI", self.model.visible_headers[2].original_header)
        self.assertTrue("NUMERO", self.model.visible_headers[1].original_header)
        self.model.move_down(1)
        self.assertTrue("SUKUPUOLI", self.model.visible_headers[1].original_header)
        self.assertTrue("NUMERO", self.model.visible_headers[2].original_header)
        self.assertEqual(1000, self.model.visible_headers[1].index)
        self.assertEqual(2000, self.model.visible_headers[2].index)

    def test_move_up(self):
        self.model.visible_headers[1].index = 1000
        self.model.visible_headers[2].index = 2000
        self.assertTrue("SUKUPUOLI", self.model.visible_headers[2].original_header)
        self.assertTrue("NUMERO", self.model.visible_headers[1].original_header)
        self.model.move_up(2)
        self.assertTrue("SUKUPUOLI", self.model.visible_headers[1].original_header)
        self.assertTrue("NUMERO", self.model.visible_headers[2].original_header)
        self.assertEqual(1000, self.model.visible_headers[1].index)
        self.assertEqual(2000, self.model.visible_headers[2].index)

    @mock.patch('kvalikirstu2.kvali_interface.rename_daf_files')
    def test_rename(self, _mock):
        self.model.rename_daf_files()
        self.assertTrue(_mock.called)

    def test_undo_redo(self):
        self.model.set_row_name(1, "aaa")
        self.assertEqual("aaa", self.model.get_headerline(1).edited_header)
        self.model.undo()
        self.assertEqual("NUMERO", self.model.get_headerline(1).edited_header)
        self.model.redo()
        self.assertEqual("aaa", self.model.get_headerline(1).edited_header)

    def test_load_backup(self):
        argument_parser.change_setting('max_backup_folder_size', 80000)
        argument_parser.change_setting('max_backups', 15)
        kvali_interface._backup_study()
        self.assertEqual("NUMERO", self.model.get_headerline(1).edited_header)
        self.model.set_row_name(1, "aaa")
        self.assertEqual("aaa", self.model.get_headerline(1).edited_header)
        self.model.restore_backup(0)
        self.assertEqual("NUMERO", self.model.get_headerline(1).edited_header)

    def test_show_type(self):
        self.assertEqual(10, len(self.model.visible_headers))
        self.model.set_type_shown(gui_model.HeaderType.BUILTIN, False)
        self.assertEqual(5, len(self.model.visible_headers))

    def test_select_language(self):
        self.model.select_index_language('test')
        self.assertEqual('test', argument_parser.get_args().index_lang)

    @mock.patch('kvalikirstu2.kvali_interface.setup_external_file_study')
    def test_setup_external_study(self, _mock):
        self.model.setup_external_file_study(['1', '2', '3'])
        _mock.assert_called_with(['1', '2', '3'])

    def test_not_enough_headers_selected(self):
        self.model.change_header_property([1, 2, 3, 4],
                                          lambda header: header.set_status(gui_model.HeaderStatus.NOT_HEADER))
        with self.assertRaises(exceptions.HeaderEditException):
            self.model.save_changes()

    def test_set_daf_header_multiple_times(self):
        self.model.change_header_property([1, 2], lambda header: header.set_status(gui_model.HeaderStatus.DAF_HEADER))
        self.model.check_changes_valid()

    @mock.patch('kvalikirstu2.converter.convert_file')
    def test_convert_encoding_called(self, mock):
        self.model.convert_txt_encoding()
        self.assertTrue(mock.called)
